# Copyright (c) 2013-2017 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module


class Git:
    commit = None
    if lib.gitCommit and lib.gitCommit != "(unknown)":
        commit = ffi.string(lib.gitCommit).decode('utf-8')

    commitShort = None
    if lib.gitCommitShort and lib.gitCommitShort != "(unknown)":
        commitShort = ffi.string(lib.gitCommitShort).decode('utf-8')

    branch = None
    if lib.gitBranch and lib.gitBranch != "(unknown)":
        branch = ffi.string(lib.gitBranch).decode('utf-8')

    revision = None
    if lib.gitRevision > 0:
        revision = lib.gitRevision


def create_callback(struct_name, cb_name, func_name=None):
    func_name = func_name or "_py{}{}".format(struct_name, cb_name[0].upper() + cb_name[1:])
    full_struct = "struct {}*".format(struct_name)

    def callback(handle, *args):
        handle = ffi.cast(full_struct, handle)
        return getattr(ffi.from_handle(handle.pyobj), cb_name)(*args)

    return ffi.def_extern(name=func_name)(callback)


__version__ = ffi.string(lib.projectVersion).decode('utf-8')



# Copyright (c) 2013-2017 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from .core import IRunner, ICoreOwner, Core
import sys


class DebuggerCoreOwner(ICoreOwner):
    def __init__(self, debugger):
        self.debugger = debugger
        self.was_paused = False

    def claim(self):
        if self.debugger.isRunning():
            self.was_paused = True
            self.debugger.pause()
        return self.debugger._core

    def release(self):
        if self.was_paused:
            self.debugger.unpause()


class NativeDebugger(IRunner):
    WATCHPOINT_WRITE = lib.WATCHPOINT_WRITE
    WATCHPOINT_READ = lib.WATCHPOINT_READ
    WATCHPOINT_RW = lib.WATCHPOINT_RW

    BREAKPOINT_HARDWARE = lib.BREAKPOINT_HARDWARE
    BREAKPOINT_SOFTWARE = lib.BREAKPOINT_SOFTWARE

    ENTER_MANUAL = lib.DEBUGGER_ENTER_MANUAL
    ENTER_ATTACHED = lib.DEBUGGER_ENTER_ATTACHED
    ENTER_BREAKPOINT = lib.DEBUGGER_ENTER_BREAKPOINT
    ENTER_WATCHPOINT = lib.DEBUGGER_ENTER_WATCHPOINT
    ENTER_ILLEGAL_OP = lib.DEBUGGER_ENTER_ILLEGAL_OP

    def __init__(self, native):
        self._native = native
        self._cbs = []
        self._core = Core._detect(native.core)
        self._core._load()

    def pause(self):
        lib.mDebuggerEnter(self._native, lib.DEBUGGER_ENTER_MANUAL, ffi.NULL)

    def unpause(self):
        self._native.state = lib.DEBUGGER_RUNNING

    @property
    def running(self):
        return self._native.state == lib.DEBUGGER_RUNNING

    @property
    def paused(self):
        return self._native.state in (lib.DEBUGGER_PAUSED, lib.DEBUGGER_CALLBACK)

    def use_core(self):
        return DebuggerCoreOwner(self)

    def set_breakpoint(self, address):
        if not self._native.platform.setBreakpoint:
            raise RuntimeError("Platform does not support breakpoints")
        self._native.platform.setBreakpoint(self._native.platform, address)

    def clear_breakpoint(self, address):
        if not self._native.platform.setBreakpoint:
            raise RuntimeError("Platform does not support breakpoints")
        self._native.platform.clearBreakpoint(self._native.platform, address)

    def set_watchpoint(self, address):
        if not self._native.platform.setWatchpoint:
            raise RuntimeError("Platform does not support watchpoints")
        self._native.platform.setWatchpoint(self._native.platform, address)

    def clear_watchpoint(self, address):
        if not self._native.platform.clearWatchpoint:
            raise RuntimeError("Platform does not support watchpoints")
        self._native.platform.clearWatchpoint(self._native.platform, address)

    def add_callback(self, callback):
        self._cbs.append(callback)


class CLIBackend(object):
    def __init__(self, backend):
        self.backend = backend

    def write(self, string):
        self.backend.printf(string)


class CLIDebugger(NativeDebugger):
    def __init__(self, native):
        super(CLIDebugger, self).__init__(native)
        self._cli = ffi.cast("struct CLIDebugger*", native)

    def printf(self, message, *args, **kwargs):
        message = message.format(*args, **kwargs)
        self._cli.backend.printf(ffi.new("char []", b"%s"), ffi.new("char []", message.encode('utf-8')))

    def install_print(self):
        sys.stdout = CLIBackend(self)



# Copyright (c) 2013-2017 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
try:
    import mgba_gamedata
except ImportError:
    pass


def search(core):
    crc32 = None
    if hasattr(core, 'PLATFORM_GBA') and core.platform == core.PLATFORM_GBA:
        platform = 'GBA'
        crc32 = core.crc32
    if hasattr(core, 'PLATFORM_GB') and core.platform == core.PLATFORM_GB:
        platform = 'GB'
        crc32 = core.crc32
    cls = mgba_gamedata.registry.search(platform, {'crc32': crc32})
    if not cls:
        return None
    return cls(core.memory.u8)



import errno
import itertools
import os
import os.path
import pytest
from configparser import ConfigParser

def pytest_addoption(parser):
    parser.addoption("--rebaseline", action="store_true", help="output a new baseline instead of testing")
    parser.addoption("--mark-failing", action="store_true", help="mark all failing tests as failing")
    parser.addoption("--mark-succeeding", action="store_true", help="unmark all succeeding tests marked as failing")
    parser.addoption("--output-diff", help="output diffs for failed tests to directory")

EXPECTED = 'expected_%04u.png'
RESULT = 'result_%04u.png'
DIFF = 'diff_%04u.png'
DIFF_NORM = 'diff_norm_%04u.png'

def pytest_exception_interact(node, call, report):
    outroot = node.config.getoption("--output-diff")
    if report.failed and hasattr(node, 'funcargs'):
        vtest = node.funcargs.get('vtest')
        if outroot:
            if not vtest:
                return
            outdir = os.path.join(outroot, *vtest.full_path)
            try:
                os.makedirs(outdir)
            except OSError as e:
                if e.errno == errno.EEXIST and os.path.isdir(outdir):
                    pass
                else:
                    raise
            for i, expected, result, diff, diffNorm in zip(itertools.count(), vtest.baseline, vtest.frames, *zip(*vtest.diffs)):
                result.save(os.path.join(outdir, RESULT % i))
                if expected:
                    expected.save(os.path.join(outdir, EXPECTED % i))
                    diff.save(os.path.join(outdir, DIFF % i))
                    diffNorm.save(os.path.join(outdir, DIFF_NORM % i))

        if node.config.getoption("--mark-failing"):
            settings = ConfigParser()
            try:
                with open(os.path.join(vtest.path, 'config.ini'), 'r') as f:
                    settings.read_file(f)
            except IOError:
                pass
            settings.set('testinfo', 'fail', '1')
            with open(os.path.join(vtest.path, 'config.ini'), 'w') as f:
                settings.write(f)



import cffi
import os, os.path
import shlex
import subprocess
import sys

ffi = cffi.FFI()
pydir = os.path.dirname(os.path.abspath(__file__))
srcdir = os.path.join(pydir, "..", "..")
incdir = os.path.join(pydir, "..", "..", "..", "include")
bindir = os.environ.get("BINDIR", os.path.join(os.getcwd(), ".."))
libdir = os.environ.get("LIBDIR")

cpp = shlex.split(os.environ.get("CPP", "cc -E"))
cppflags = shlex.split(os.environ.get("CPPFLAGS", ""))
cppflags.extend(["-I" + incdir, "-I" + srcdir, "-I" + bindir])

ffi.set_source("mgba._pylib", """
#define static
#define inline
#define MGBA_EXPORT
#include <mgba/flags.h>
#define OPAQUE_THREADING
#include <mgba/core/blip_buf.h>
#include <mgba/core/cache-set.h>
#include <mgba-util/common.h>
#include <mgba/core/core.h>
#include <mgba/core/map-cache.h>
#include <mgba/core/log.h>
#include <mgba/core/mem-search.h>
#include <mgba/core/thread.h>
#include <mgba/core/version.h>
#include <mgba/debugger/debugger.h>
#include <mgba/gba/interface.h>
#include <mgba/internal/arm/arm.h>
#include <mgba/internal/debugger/cli-debugger.h>
#include <mgba/internal/gba/gba.h>
#include <mgba/internal/gba/input.h>
#include <mgba/internal/gba/renderers/cache-set.h>
#include <mgba/internal/sm83/sm83.h>
#include <mgba/internal/gb/gb.h>
#include <mgba/internal/gb/renderers/cache-set.h>
#include <mgba-util/png-io.h>
#include <mgba-util/vfs.h>

#define PYEXPORT
#include "platform/python/core.h"
#include "platform/python/log.h"
#include "platform/python/sio.h"
#include "platform/python/vfs-py.h"
#undef PYEXPORT
""", include_dirs=[incdir, srcdir],
     extra_compile_args=cppflags,
     libraries=["mgba"],
     library_dirs=[bindir],
     runtime_library_dirs=[libdir],
     sources=[os.path.join(pydir, path) for path in ["vfs-py.c", "core.c", "log.c", "sio.c"]])

preprocessed = subprocess.check_output(cpp + ["-fno-inline", "-P"] + cppflags + [os.path.join(pydir, "_builder.h")], universal_newlines=True)

lines = []
for line in preprocessed.splitlines():
    line = line.strip()
    if line.startswith('#'):
        continue
    lines.append(line)
ffi.cdef('\n'.join(lines))

preprocessed = subprocess.check_output(cpp + ["-fno-inline", "-P"] + cppflags + [os.path.join(pydir, "lib.h")], universal_newlines=True)

lines = []
for line in preprocessed.splitlines():
    line = line.strip()
    if line.startswith('#'):
        continue
    lines.append(line)
ffi.embedding_api('\n'.join(lines))

ffi.embedding_init_code("""
    import os, os.path
    from mgba._pylib import ffi, lib
    symbols = {}
    globalSyms = {
        'symbols': symbols
    }
    pendingCode = []

    @ffi.def_extern()
    def mPythonSetDebugger(debugger):
        from mgba.debugger import NativeDebugger, CLIDebugger
        oldDebugger = globalSyms.get('debugger')
        if oldDebugger and oldDebugger._native == debugger:
            return
        if oldDebugger and not debugger:
            del globalSyms['debugger']
            return
        if debugger.type == lib.DEBUGGER_CLI:
            debugger = CLIDebugger(debugger)
        else:
            debugger = NativeDebugger(debugger)
        globalSyms['debugger'] = debugger

    @ffi.def_extern()
    def mPythonLoadScript(name, vf):
        from mgba.vfs import VFile
        vf = VFile(vf)
        name = ffi.string(name)
        source = vf.read_all().decode('utf-8')
        try:
            code = compile(source, name, 'exec')
            pendingCode.append(code)
        except:
            return False
        return True

    @ffi.def_extern()
    def mPythonRunPending():
        global pendingCode
        for code in pendingCode:
            exec(code, globalSyms, {})
        pendingCode = []

    @ffi.def_extern()
    def mPythonDebuggerEntered(reason, info):
        debugger = globalSyms['debugger']
        if not debugger:
            return
        if info == ffi.NULL:
            info = None
        for cb in debugger._cbs:
            cb(reason, info)

    @ffi.def_extern()
    def mPythonLookupSymbol(name, outptr):
        name = ffi.string(name).decode('utf-8')
        if name not in symbols:
            return False
        sym = symbols[name]
        val = None
        try:
            val = int(sym)
        except:
            try:
                val = sym()
            except:
                pass
        if val is None:
            return False
        try:
            outptr[0] = ffi.cast('int32_t', val)
            return True
        except:
            return False
""")

if __name__ == "__main__":
    ffi.emit_c_code("lib.c")



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module


class _ARMRegisters:
    def __init__(self, cpu):
        self._cpu = cpu

    def __getitem__(self, reg):
        if reg > lib.ARM_PC:
            raise IndexError("Register out of range")
        return self._cpu._native.gprs[reg]

    def __setitem__(self, reg, value):
        if reg >= lib.ARM_PC:
            raise IndexError("Register out of range")
        self._cpu._native.gprs[reg] = value


class ARMCore:
    SP = 13
    LR = 14
    PC = 15

    def __init__(self, native):
        self._native = ffi.cast("struct ARMCore*", native)
        self.gprs = _ARMRegisters(self)
        self.cpsr = self._native.cpsr
        self.spsr = self._native.spsr

    @property
    def sp(self):
        return self.gprs[self.SP]

    @sp.setter
    def sp(self, value):
        self.gprs[self.SP] = value

    @property
    def lr(self):
        return self.gprs[self.LR]

    @lr.setter
    def lr(self, value):
        self.gprs[self.LR] = value

    @property
    def pc(self):
        return self.gprs[self.PC]

    @pc.setter
    def pc(self, value):
        self.gprs[self.PC] = value



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from . import tile, audio
from cached_property import cached_property
from functools import wraps


def find(path):
    core = lib.mCoreFind(path.encode('UTF-8'))
    if core == ffi.NULL:
        return None
    return Core._init(core)


def find_vf(vfile):
    core = lib.mCoreFindVF(vfile.handle)
    if core == ffi.NULL:
        return None
    return Core._init(core)


def load_path(path):
    core = find(path)
    if not core or not core.load_file(path):
        return None
    return core


def load_vf(vfile):
    core = find_vf(vfile)
    if not core or not core.load_rom(vfile):
        return None
    return core


def needs_reset(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self._was_reset:
            raise RuntimeError("Core must be reset first")
        return func(self, *args, **kwargs)
    return wrapper


def protected(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self._protected:
            raise RuntimeError("Core is protected")
        return func(self, *args, **kwargs)
    return wrapper


@ffi.def_extern()
def _mCorePythonCallbacksVideoFrameStarted(user):  # pylint: disable=invalid-name
    context = ffi.from_handle(user)
    context._video_frame_started()


@ffi.def_extern()
def _mCorePythonCallbacksVideoFrameEnded(user):  # pylint: disable=invalid-name
    context = ffi.from_handle(user)
    context._video_frame_ended()


@ffi.def_extern()
def _mCorePythonCallbacksCoreCrashed(user):  # pylint: disable=invalid-name
    context = ffi.from_handle(user)
    context._core_crashed()


@ffi.def_extern()
def _mCorePythonCallbacksSleep(user):  # pylint: disable=invalid-name
    context = ffi.from_handle(user)
    context._sleep()


@ffi.def_extern()
def _mCorePythonCallbacksKeysRead(user):  # pylint: disable=invalid-name
    context = ffi.from_handle(user)
    context._keys_read()


class CoreCallbacks(object):
    def __init__(self):
        self._handle = ffi.new_handle(self)
        self.video_frame_started = []
        self.video_frame_ended = []
        self.core_crashed = []
        self.sleep = []
        self.keys_read = []
        self.context = lib.mCorePythonCallbackCreate(self._handle)

    def _video_frame_started(self):
        for callback in self.video_frame_started:
            callback()

    def _video_frame_ended(self):
        for callback in self.video_frame_ended:
            callback()

    def _core_crashed(self):
        for callback in self.core_crashed:
            callback()

    def _sleep(self):
        for callback in self.sleep:
            callback()

    def _keys_read(self):
        for callback in self.keys_read:
            callback()


class Core(object):
    if hasattr(lib, 'mPLATFORM_GBA'):
        PLATFORM_GBA = lib.mPLATFORM_GBA

    if hasattr(lib, 'mPLATFORM_GB'):
        PLATFORM_GB = lib.mPLATFORM_GB

    def __init__(self, native):
        self._core = native
        self._was_reset = False
        self._protected = False
        self._callbacks = CoreCallbacks()
        self._core.addCoreCallbacks(self._core, self._callbacks.context)
        self.config = Config(ffi.addressof(native.config))

    def __del__(self):
        self._was_reset = False

    @cached_property
    def graphics_cache(self):
        if not self._was_reset:
            raise RuntimeError("Core must be reset first")
        return tile.CacheSet(self)

    @cached_property
    def tiles(self):
        tiles = []
        native_tiles = ffi.addressof(self.graphics_cache.cache.tiles)
        for i in range(lib.mTileCacheSetSize(native_tiles)):
            tiles.append(tile.TileView(lib.mTileCacheSetGetPointer(native_tiles, i)))
        return tiles

    @cached_property
    def maps(self):
        maps = []
        native_maps = ffi.addressof(self.graphics_cache.cache.maps)
        for i in range(lib.mMapCacheSetSize(native_maps)):
            maps.append(tile.MapView(lib.mMapCacheSetGetPointer(native_maps, i)))
        return maps

    @classmethod
    def _init(cls, native):
        core = ffi.gc(native, native.deinit)
        success = bool(core.init(core))
        lib.mCoreInitConfig(core, ffi.NULL)
        if not success:
            raise RuntimeError("Failed to initialize core")
        return cls._detect(core)

    @classmethod
    def _detect(cls, core):
        if hasattr(cls, 'PLATFORM_GBA') and core.platform(core) == cls.PLATFORM_GBA:
            from .gba import GBA
            return GBA(core)
        if hasattr(cls, 'PLATFORM_GB') and core.platform(core) == cls.PLATFORM_GB:
            from .gb import GB
            return GB(core)
        return Core(core)

    def _load(self):
        self._was_reset = True

    @protected
    def load_file(self, path):
        return bool(lib.mCoreLoadFile(self._core, path.encode('UTF-8')))

    def is_rom(self, vfile):
        return bool(self._core.isROM(vfile.handle))

    @protected
    def load_rom(self, vfile):
        return bool(self._core.loadROM(self._core, vfile.handle))

    @protected
    def load_bios(self, vfile, id=0):
        res = bool(self._core.loadBIOS(self._core, vfile.handle, id))
        if res:
            vfile._claimed = True
        return res

    @protected
    def load_save(self, vfile):
        res = bool(self._core.loadSave(self._core, vfile.handle))
        if res:
            vfile._claimed = True
        return res

    @protected
    def load_temporary_save(self, vfile):
        return bool(self._core.loadTemporarySave(self._core, vfile.handle))

    @protected
    def load_patch(self, vfile):
        return bool(self._core.loadPatch(self._core, vfile.handle))

    @protected
    def load_config(self, config):
        lib.mCoreLoadForeignConfig(self._core, config._native)

    @protected
    def autoload_save(self):
        return bool(lib.mCoreAutoloadSave(self._core))

    @protected
    def autoload_patch(self):
        return bool(lib.mCoreAutoloadPatch(self._core))

    @protected
    def autoload_cheats(self):
        return bool(lib.mCoreAutoloadCheats(self._core))

    @property
    def platform(self):
        return self._core.platform(self._core)

    @protected
    def desired_video_dimensions(self):
        width = ffi.new("unsigned*")
        height = ffi.new("unsigned*")
        self._core.desiredVideoDimensions(self._core, width, height)
        return width[0], height[0]

    @protected
    def set_video_buffer(self, image):
        self._core.setVideoBuffer(self._core, image.buffer, image.stride)

    @protected
    def set_audio_buffer_size(self, size):
        self._core.setAudioBufferSize(self._core, size)

    @property
    def audio_buffer_size(self):
        return self._core.getAudioBufferSize(self._core)

    @protected
    def get_audio_channels(self):
        return audio.StereoBuffer(self.get_audio_channel(0), self.get_audio_channel(1));

    @protected
    def get_audio_channel(self, channel):
        return audio.Buffer(self._core.getAudioChannel(self._core, channel), self.frequency)

    @protected
    def reset(self):
        self._core.reset(self._core)
        self._load()

    @needs_reset
    @protected
    def run_frame(self):
        self._core.runFrame(self._core)

    @needs_reset
    @protected
    def run_loop(self):
        self._core.runLoop(self._core)

    @needs_reset
    @protected
    def step(self):
        self._core.step(self._core)

    @needs_reset
    @protected
    def load_raw_state(self, state):
        if len(state) < self._core.stateSize(self._core):
            return False
        return self._core.loadState(self._core, state)

    @needs_reset
    @protected
    def save_raw_state(self):
        state = ffi.new('unsigned char[%i]' % self._core.stateSize(self._core))
        if self._core.saveState(self._core, state):
            return state
        return None

    @staticmethod
    def _keys_to_int(*args, **kwargs):
        keys = 0
        if 'raw' in kwargs:
            keys = kwargs['raw']
        for key in args:
            keys |= 1 << key
        return keys

    @protected
    def set_keys(self, *args, **kwargs):
        self._core.setKeys(self._core, self._keys_to_int(*args, **kwargs))

    @protected
    def add_keys(self, *args, **kwargs):
        self._core.addKeys(self._core, self._keys_to_int(*args, **kwargs))

    @protected
    def clear_keys(self, *args, **kwargs):
        self._core.clearKeys(self._core, self._keys_to_int(*args, **kwargs))

    @property
    @needs_reset
    def frame_counter(self):
        return self._core.frameCounter(self._core)

    @property
    def frame_cycles(self):
        return self._core.frameCycles(self._core)

    @property
    def frequency(self):
        return self._core.frequency(self._core)

    @property
    def game_title(self):
        title = ffi.new("char[16]")
        self._core.getGameTitle(self._core, title)
        return ffi.string(title, 16).decode("ascii")

    @property
    def game_code(self):
        code = ffi.new("char[12]")
        self._core.getGameCode(self._core, code)
        return ffi.string(code, 12).decode("ascii")

    def add_frame_callback(self, callback):
        self._callbacks.video_frame_ended.append(callback)

    @property
    def crc32(self):
        return self._native.romCrc32


class ICoreOwner(object):
    def claim(self):
        raise NotImplementedError

    def release(self):
        raise NotImplementedError

    def __enter__(self):
        self.core = self.claim()
        self.core._protected = True
        return self.core

    def __exit__(self, type, value, traceback):
        self.core._protected = False
        self.release()


class IRunner(object):
    def pause(self):
        raise NotImplementedError

    def unpause(self):
        raise NotImplementedError

    def use_core(self):
        raise NotImplementedError

    @property
    def running(self):
        raise NotImplementedError

    @property
    def paused(self):
        raise NotImplementedError


class Config(object):
    def __init__(self, native=None, port=None, defaults={}):
        if not native:
            self._port = ffi.NULL
            if port:
                self._port = ffi.new("char[]", port.encode("UTF-8"))
            native = ffi.gc(ffi.new("struct mCoreConfig*"), lib.mCoreConfigDeinit)
            lib.mCoreConfigInit(native, self._port)
        self._native = native
        for key, value in defaults.items():
            if isinstance(value, bool):
                value = int(value)
            lib.mCoreConfigSetDefaultValue(self._native, ffi.new("char[]", key.encode("UTF-8")), ffi.new("char[]", str(value).encode("UTF-8")))

    def __getitem__(self, key):
        string = lib.mCoreConfigGetValue(self._native, ffi.new("char[]", key.encode("UTF-8")))
        if not string:
            return None
        return ffi.string(string)

    def __setitem__(self, key, value):
        if isinstance(value, bool):
            value = int(value)
        lib.mCoreConfigSetValue(self._native, ffi.new("char[]", key.encode("UTF-8")), ffi.new("char[]", str(value).encode("UTF-8")))



import pytest
import cinema.test
import mgba.log
import os.path
import yaml

mgba.log.install_default(mgba.log.NullLogger())

def flatten(d):
    l = []
    for k, v in d.tests.items():
        if v.tests:
            l.extend(flatten(v))
        else:
            l.append(v)
    l.sort()
    return l

def pytest_generate_tests(metafunc):
    if 'vtest' in metafunc.fixturenames:
        tests = cinema.test.gather_tests(os.path.join(os.path.dirname(__file__), '..', '..', '..', 'cinema'))
        testList = flatten(tests)
        params = []
        for test in testList:
            marks = []
            xfail = test.settings.get('fail')
            if xfail and bool(xfail):
                marks = pytest.mark.xfail(reason=xfail if isinstance(xfail, str) else None)
            params.append(pytest.param(test, id=test.name, marks=marks))
        metafunc.parametrize('vtest', params, indirect=True)

@pytest.fixture
def vtest(request):
    return request.param

def test_video(vtest, pytestconfig):
    vtest.setup()
    if pytestconfig.getoption('--rebaseline'):
        vtest.generate_baseline()
    else:
        try:
            vtest.test()
        except IOError:
            raise
        if pytestconfig.getoption('--mark-succeeding') and 'fail' in vtest.settings:
            # TODO: This can fail if an entire directory is marked as failing
            settings = {}
            try:
                with open(os.path.join(vtest.path, 'manifest.yml'), 'r') as f:
                    settings = yaml.safe_load(f)
            except IOError:
                pass
            if 'fail' in settings:
                del settings['fail']
            else:
                settings['fail'] = False
            if settings:
                with open(os.path.join(vtest.path, 'manifest.yml'), 'w') as f:
                    yaml.dump(settings, f, default_flow_style=False)
            else:
                os.remove(os.path.join(vtest.path, 'manifest.yml'))



import pytest
import os

import mgba.vfs as vfs
from mgba._pylib import ffi


def test_vfs_open():
    with open(__file__) as f:
        vf = vfs.open(f)
        assert vf
        assert vf.close()


def test_vfs_open_path():
    vf = vfs.open_path(__file__)
    assert vf
    assert vf.close()


def test_vfs_read():
    vf = vfs.open_path(__file__)
    buffer = ffi.new('char[13]')
    assert vf.read(buffer, 13) == 13
    assert ffi.string(buffer) == b'import pytest'
    vf.close()


def test_vfs_readline():
    vf = vfs.open_path(__file__)
    buffer = ffi.new('char[16]')
    linelen = vf.readline(buffer, 16)
    assert linelen in (14, 15)
    if linelen == 14:
        assert ffi.string(buffer) == b'import pytest\n'
    elif linelen == 15:
        assert ffi.string(buffer) == b'import pytest\r\n'
    vf.close()


def test_vfs_read_all_size():
    vf = vfs.open_path(__file__)
    buffer = vf.read_all()
    assert buffer
    assert len(buffer)
    assert len(buffer) == vf.size()
    vf.close()


def test_vfs_seek():
    vf = vfs.open_path(__file__)
    assert vf.seek(0, os.SEEK_SET) == 0
    assert vf.seek(1, os.SEEK_SET) == 1
    assert vf.seek(1, os.SEEK_CUR) == 2
    assert vf.seek(-1, os.SEEK_CUR) == 1
    assert vf.seek(0, os.SEEK_CUR) == 1
    assert vf.seek(0, os.SEEK_END) == vf.size()
    assert vf.seek(-1, os.SEEK_END) == vf.size() -1
    vf.close()


def test_vfs_open_path_invalid():
    vf = vfs.open_path('.invalid')
    assert not vf



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi  # pylint: disable=no-name-in-module


class SM83Core:
    # pylint: disable=invalid-name
    def __init__(self, native):
        self._native = ffi.cast("struct SM83Core*", native)

    @property
    def a(self):
        return self._native.a

    @property
    def b(self):
        return self._native.b

    @property
    def c(self):
        return self._native.c

    @property
    def d(self):
        return self._native.d

    @property
    def e(self):
        return self._native.e

    @property
    def f(self):
        return self._native.f

    @property
    def h(self):
        return self._native.h

    @property
    def l(self):
        return self._native.l

    @property
    def sp(self):
        return self._native.sp

    @property
    def pc(self):
        return self._native.pc

    @property
    def af(self):
        return (self.a << 8) | self.f

    @property
    def bc(self):
        return (self.b << 8) | self.c

    @property
    def de(self):
        return (self.d << 8) | self.e

    @property
    def hl(self):
        return (self.h << 8) | self.l

    @a.setter
    def a(self, value):
        self._native.a = value

    @b.setter
    def b(self, value):
        self._native.b = value

    @c.setter
    def c(self, value):
        self._native.c = value

    @d.setter
    def d(self, value):
        self._native.d = value

    @e.setter
    def e(self, value):
        self._native.e = value

    @f.setter
    def f(self, value):
        self._native.f.packed = value
        self._native.f.unused = 0

    @h.setter
    def h(self, value):
        self._native.h = value

    @l.setter
    def l(self, value):
        self._native.l = value

    @sp.setter
    def sp(self, value):
        self._native.sp = value

    @af.setter
    def af(self, value):
        self.a = value >> 8
        self.f = value & 0xFF

    @bc.setter
    def bc(self, value):
        self.b = value >> 8
        self.c = value & 0xFF

    @de.setter
    def de(self, value):
        self.d = value >> 8
        self.e = value & 0xFF

    @hl.setter
    def hl(self, value):
        self.h = value >> 8
        self.l = value & 0xFF



from PIL.ImageChops import difference
from PIL.ImageOps import autocontrast
from PIL.Image import open as PIOpen


class VideoFrame(object):
    def __init__(self, pil_image):
        self.image = pil_image.convert('RGB')

    @staticmethod
    def diff(a, b):
        diff = difference(a.image, b.image)
        diff_normalized = autocontrast(diff)
        return (VideoFrame(diff), VideoFrame(diff_normalized))

    @staticmethod
    def load(path):
        with open(path, 'rb') as f:
            image = PIOpen(f)
            image.load()
            return VideoFrame(image)

    def save(self, path):
        return self.image.save(path)



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from . import create_callback

create_callback("mLoggerPy", "log", "_pyLog")


def install_default(logger):
    Logger.install_default(logger)


def silence():
    Logger.install_default(NullLogger())


class Logger(object):
    FATAL = lib.mLOG_FATAL
    DEBUG = lib.mLOG_DEBUG
    INFO = lib.mLOG_INFO
    WARN = lib.mLOG_WARN
    ERROR = lib.mLOG_ERROR
    STUB = lib.mLOG_STUB
    GAME_ERROR = lib.mLOG_GAME_ERROR

    _DEFAULT_LOGGER = None

    def __init__(self):
        self._handle = ffi.new_handle(self)
        self._native = ffi.gc(lib.mLoggerPythonCreate(self._handle), lib.free)

    @staticmethod
    def category_name(category):
        return ffi.string(lib.mLogCategoryName(category)).decode('UTF-8')

    @classmethod
    def install_default(cls, logger):
        cls._DEFAULT_LOGGER = logger
        lib.mLogSetDefaultLogger(logger._native)

    def log(self, category, level, message):
        print("{}: {}".format(self.category_name(category), message))


class NullLogger(Logger):
    def log(self, category, level, message):
        pass
